from typing import Union, Callable, List, Dict, Tuple, Type, Any, TypeVar, Optional
from pydantic import BaseModel
from requests import Response


class HTTPResponse(BaseModel):
    status_code: int = 200


class MessageResponse(HTTPResponse):
    message: str


class ErrorResponse(MessageResponse):
    pass


class ErrorHTTP403Response(ErrorResponse):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data, status_code=403, message=data['detail'])


class ErrorHTTP404Response(ErrorResponse):
    def __init__(self, **data: Any) -> None:
        if 'detail' in data:
            super().__init__(**data, status_code=404, message=data['detail'])
        else:
            super().__init__(**data, status_code=404)


class ErrorHTTP409Response(ErrorResponse):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data, status_code=409)


class ErrorHTTP422Response(ErrorResponse):
    detail: dict

    def __init__(self, **data: Any) -> None:
        super().__init__(**data, status_code=422, message="Parameters failed validation")


class ErrorHTTP500Response(ErrorResponse):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data, status_code=500, message="Server Error")


GenericHTTPResponse = TypeVar('GenericHTTPResponse', bound=HTTPResponse)


def transform_response_to_schema(response: Response, schema: Union[None, Type[HTTPResponse]]) -> GenericHTTPResponse:
    if response.status_code == 200:
        if schema:
            return schema(**response.json())
        return HTTPResponse()

    if response.status_code == 403:
        return ErrorHTTP403Response(**response.json())

    if response.status_code == 404:
        return ErrorHTTP404Response(**response.json())

    if response.status_code == 409:
        return ErrorHTTP409Response(**response.json())

    if response.status_code == 422:
        return ErrorHTTP422Response(**response.json())

    if response.status_code == 500:
        return ErrorHTTP500Response()

    return ErrorResponse(message="Undefined Error", status_code=1000)


def is_response_error(response: Union[Response, HTTPResponse]) -> bool:
    if response.status_code != 200:
        return True
    return False


class Pagination(BaseModel):
    """
    Pagination state
    """
    current_page: Optional[int] = None
    page_size: int = 25
    total_items: Optional[int] = None
    total_pages: Optional[int] = None
    has_previous_page: Optional[bool] = None
    has_next_page: Optional[bool] = True
    previous_page: Optional[int] = None
    next_page: Optional[int] = 1




class PageResponse(HTTPResponse):
    """
    Page state
    """
    items: list
    pagination: Pagination

    def transform(self, transformer: Callable):
        """
        Transforms results from a DB backbone into a Pydantic (or other) object type

        Args:
            transformer (Callable): A function which takes in an input and returns a transformed output

        Returns:
            None

        """
        items: list = self.items
        self.items = []
        for item in items:
            self.items.append(transformer(item))

    def transform_dict(self, pydantic_type: Type[BaseModel]):
        """
        Transforms results from a dict into a Pydantic (or other) object type

        Returns:
            None

        """
        items: list = self.items
        self.items = []
        for item in items:
            self.items.append(pydantic_type(**item))



from typing import Union

from pydantic import BaseModel
import requests


class SDKConfiguration(BaseModel):
    access_token: str
    secret_token: str
    project_identifier: str
    base_url_override: Union[None, str] = None


def make_requests(sdk_configuration: SDKConfiguration) -> requests.Session:
    requests_session: requests.Session = requests.Session()

    requests_session.headers.update({
        'X-Access-Token': sdk_configuration.access_token,
        'X-Secret-Token': sdk_configuration.secret_token,
        'Ehelply-Project': sdk_configuration.project_identifier

    })

    return requests_session



from pathlib import Path
import os.path as OSPath
from typing import List
import re
import os

import pdoc
from pdoc.cli import _open_write_file

import typer


cli = typer.Typer()
dev_cli = typer.Typer()

cli.add_typer(dev_cli, name="dev")


@dev_cli.command()
def export_code_docs():
    output_dir: Path = Path(__file__).parents[1].joinpath('docs')

    modules = ['ehelply_python_sdk']  # Public submodules are auto-imported
    context = pdoc.Context()

    modules = [pdoc.Module(mod, context=context)
               for mod in modules]
    pdoc.link_inheritance(context)

    def recursive_htmls(mod):
        yield mod.name, mod.html(), mod.url()
        for submod in mod.submodules():
            yield from recursive_htmls(submod)

    for mod in modules:

        for module_name, html, url in recursive_htmls(mod):
            module_name: str
            url: str
            html: str

            url_components: List[str] = url.split("/")
            url_components.pop(0)
            url: str = '/'.join(url_components)

            filepath: str = OSPath.join(str(output_dir), *re.sub(r'\.html$', '.html', url).split('/'))

            dirpath = OSPath.dirname(filepath)
            if not os.access(dirpath, os.R_OK):
                os.makedirs(dirpath)

            with _open_write_file(filepath) as f:
                f.write(html)





def cli_main():
    cli()


if __name__ == '__main__':
    cli()


from ehelply_python_sdk.services.service_schemas import *


class CreateKey(BaseModel):
    name: str
    summary: str


class CreateKeyResponse(HTTPResponse):
    uuid: str
    access: str
    secret: str


class VerifyKeyResponse(HTTPResponse):
    uuid: str
    access: str
    name: str
    summary: str
    created_at: str
    last_used_at: str


class CreateTokenResponse(HTTPResponse):
    token: str



from .schemas import *
from ehelply_python_sdk.services.service_sdk_base import SDKBase


class AuthModel:
    def __init__(
            self,
            access_sdk=None,
            active_participant_uuid=None,
            entity_identifier=None,
            project_uuid=None,
            access_token=None,
            secret_token=None,
            claims=None,
    ) -> None:
        self.access_sdk = access_sdk
        self.active_participant_uuid = active_participant_uuid
        self.entity_identifier = entity_identifier
        self.project_uuid = project_uuid
        self.access_token = access_token
        self.secret_token = secret_token
        self.claims = claims


class AccessSDK(SDKBase):
    def get_base_url(self) -> str:
        return super().get_base_url() + "/access/partitions/" + self.sdk_configuration.project_identifier

    def get_docs_url(self) -> str:
        return super().get_docs_url()

    def get_service_version(self) -> str:
        return super().get_service_version()

    def make_partition_override(self, partition: str) -> str:
        components: List[str] = self.get_base_url().split("/")
        components.pop()
        components.append(partition)
        return "/".join(components)

    def search_types(self, name: str = None, pagination: Pagination = None) -> Union[GenericHTTPResponse, PageResponse]:
        params: dict = {}
        if name:
            params["name"] = name

        if pagination:
            params["page"] = pagination.next_page
            params["page_size"] = pagination.page_size

        response: PageResponse = transform_response_to_schema(
            self.requests_session.get(
                self.get_base_url() + "/permissions/types",
                params=params
            ),
            schema=PageResponse
        )

        if is_response_error(response):
            return response

        response.transform_dict(pydantic_type=SearchTypeItem)

        return response

    def create_type(self, partition_type: CreateType) -> Union[GenericHTTPResponse, CreateTypeResponse]:
        """
        Parameters
        ----------
        partition_type : CreateType
            Pydantic model which defines the parameters of the type to create

        Returns
        -------

        """
        return transform_response_to_schema(
            self.requests_session.post(
                self.get_base_url() + "/permissions/types",
                json={"partition_type": partition_type.dict()}
            ),
            schema=CreateTypeResponse
        )

    def create_node(self, partition_type_uuid: str, node: CreateNode) -> Union[GenericHTTPResponse, CreateNodeResponse]:
        return transform_response_to_schema(
            self.requests_session.post(
                self.get_base_url() + "/permissions/types/" + partition_type_uuid + "/nodes",
                json={"node": node.dict()}
            ),
            schema=CreateNodeResponse
        )

    def search_nodes(self, type_uuid: str, node: str = None, pagination: Pagination = None) -> Union[GenericHTTPResponse, PageResponse]:
        params: dict = {}
        if node:
            params["node"] = node

        if pagination:
            params["page"] = pagination.next_page
            params["page_size"] = pagination.page_size

        response: PageResponse = transform_response_to_schema(
            self.requests_session.get(
                self.get_base_url() + "/permissions/types/" + type_uuid + "/nodes",
                params=params
            ),
            schema=PageResponse
        )

        if is_response_error(response):
            return response

        response.transform_dict(pydantic_type=SearchNodeItem)

        return response

    def create_group(self, group: CreateGroup) -> Union[GenericHTTPResponse, CreateGroupResponse]:
        return transform_response_to_schema(
            self.requests_session.post(
                self.get_base_url() + "/who/groups",
                json={"group": group.dict()}
            ),
            schema=CreateGroupResponse
        )

    def create_role(self, role: CreateRole) -> Union[GenericHTTPResponse, CreateRoleResponse]:
        return transform_response_to_schema(
            self.requests_session.post(
                self.get_base_url() + "/roles",
                json={"role": role.dict()}
            ),
            schema=CreateRoleResponse
        )

    def add_node_to_role(self, node_uuid: str, role_uuid: str) -> Union[
        GenericHTTPResponse, MessageResponse]:
        return transform_response_to_schema(
            self.requests_session.post(
                self.get_base_url() + "/roles/" + role_uuid + "/nodes/" + node_uuid
            ),
            schema=MessageResponse
        )

    def remove_node_from_role(self, node_uuid: str, role_uuid: str) -> Union[
        GenericHTTPResponse, MessageResponse]:
        return transform_response_to_schema(
            self.requests_session.delete(
                self.get_base_url() + "/roles/" + role_uuid + "/nodes/" + node_uuid
            ),
            schema=MessageResponse
        )
    
    def add_node_to_key(self, node_uuid: str, key_uuid: str) -> Union[
        GenericHTTPResponse, MessageResponse]:
        return transform_response_to_schema(
            self.requests_session.post(
                self.get_base_url() + "/keys/" + key_uuid + "/nodes/" + node_uuid
            ),
            schema=MessageResponse
        )

    def remove_node_from_key(self, node_uuid: str, key_uuid: str) -> Union[
        GenericHTTPResponse, MessageResponse]:
        return transform_response_to_schema(
            self.requests_session.delete(
                self.get_base_url() + "/keys/" + key_uuid + "/nodes/" + node_uuid
            ),
            schema=MessageResponse
        )

    def add_entity_to_group(self, entity_identifier: str, group_uuid: str) -> Union[
        GenericHTTPResponse, MessageResponse]:
        return transform_response_to_schema(
            self.requests_session.post(
                self.get_base_url() + "/who/groups/" + group_uuid + "/entities/" + entity_identifier
            ),
            schema=MessageResponse
        )

    def remove_entity_from_group(self, entity_identifier: str, group_uuid: str) -> Union[
        GenericHTTPResponse, MessageResponse]:
        return transform_response_to_schema(
            self.requests_session.delete(
                self.get_base_url() + "/who/groups/" + group_uuid + "/entities/" + entity_identifier
            ),
            schema=MessageResponse
        )

    def attach_key_to_entity(
            self,
            entity_identifier: str,
            key_uuid: str
    ) -> Union[GenericHTTPResponse, AttachKeyToEntityResponse]:
        return transform_response_to_schema(
            self.requests_session.post(
                self.get_base_url() + "/who/entities/" + entity_identifier + "/keys/" + key_uuid
            ),
            schema=AttachKeyToEntityResponse
        )

    def make_rgt(
            self,
            role_uuid: str,
            group_uuid: str,
            target_identifier: str
    ) -> Union[GenericHTTPResponse, MakeRGTResponse]:
        return transform_response_to_schema(
            self.requests_session.post(
                self.get_base_url() + "/rgts/roles/" + role_uuid + "/groups/" + group_uuid + "/targets/" + target_identifier,
                json={"limits": []}
            ),
            schema=MakeRGTResponse
        )

    def automate_role_group_rgt(
            self,
            role: CreateRole,
            group: CreateGroup,
            target_identifier: str,
            entity_identifiers: List[str] = None
    ) -> Union[GenericHTTPResponse, MakeRGTResponse]:

        role_response: CreateRoleResponse = self.create_role(role=role)

        if is_response_error(role_response):
            return role_response

        group_response: CreateGroupResponse = self.create_group(group=group)

        if is_response_error(group_response):
            return group_response

        if len(entity_identifiers) > 0 and not group.default:
            for entity_identifier in entity_identifiers:
                self.add_entity_to_group(entity_identifier=entity_identifier, group_uuid=group_response.uuid)

        return self.make_rgt(
            role_uuid=role_response.uuid,
            group_uuid=group_response.uuid,
            target_identifier=target_identifier
        )

    def get_entity_for_key(
            self,
            key_uuid: str
    ) -> Union[GenericHTTPResponse, GetEntityForKeyResponse]:
        return transform_response_to_schema(
            self.requests_session.post(
                self.get_base_url() + "/who/entities/keys/" + key_uuid
            ),
            schema=GetEntityForKeyResponse
        )

    def is_entity_allowed(
            self,
            entity_identifier: str,
            target_identifier: str,
            node: str,
            partition: str = None
    ) -> bool:

        if partition:
            base_url: str = self.make_partition_override(partition=partition)
        else:
            base_url: str = self.get_base_url()

        response: Response = self.requests_session.get(
            base_url + "/auth/targets/" + target_identifier + "/nodes/" + node + "/entities/" + entity_identifier
        )

        if not is_response_error(response) and response.json() is True:
            return True

        # return transform_response_to_schema(response, None)
        return False

    def is_key_allowed(
            self,
            access_token: str,
            secret_token: str,
            target_identifier: str,
            node: str,
            partition: str = None
    ) -> bool:
        if partition:
            base_url: str = self.make_partition_override(partition=partition)
        else:
            base_url: str = self.get_base_url()

        headers: dict = {
            "X-Access-Token": access_token,
            "X-Secret-Token": secret_token
        }

        response: Response = self.requests_session.get(
            base_url + "/auth/targets/" + target_identifier + "/nodes/" + node + "/keys",
            headers=headers
        )

        if not is_response_error(response) and response.json() is True:
            return True

        # return transform_response_to_schema(response, None)
        return False

    def is_allowed(
            self,
            auth_model: AuthModel,
            target_identifier: str,
            node: str,
            partition: str = None
    ) -> bool:
        if auth_model.entity_identifier:
            if self.is_entity_allowed(
                    entity_identifier=auth_model.entity_identifier,
                    target_identifier=target_identifier,
                    node=node,
                    partition=partition,
            ):
                return True

        if auth_model.access_token and auth_model.secret_token:
            if self.is_key_allowed(
                    access_token=auth_model.access_token,
                    secret_token=auth_model.secret_token,
                    target_identifier=target_identifier,
                    node=node,
                    partition=partition,
            ):
                return True

        return False

    # def templated_access(
    #         self,
    #         auth_model: AuthModel,
    #         service_target_identifier: str,
    #         target_identifier: str,
    #         node: str
    # ):
    #     AuthRule(
    #         auth_model,
    #         AuthRule(auth_model).participant_has_node_on_target(
    #             partition='ehelply-resources',
    #             node=node,
    #             target_identifier=service_target_identifier
    #         ),
    #         AuthRule(auth_model).participant_has_node_on_target(
    #             partition='ehelply-cloud',
    #             node=node,
    #             target_identifier=auth_model.project_uuid
    #         ).customentity_has_node_on_target(
    #             partition='ehelply-cloud',
    #             node=node,
    #             target_identifier=service_target_identifier,
    #             entity_identifier=auth_model.project_uuid
    #         ).entity_has_node_on_target(
    #             node=node,
    #             target_identifier=target_identifier,
    #             partition=auth_model.project_uuid
    #         )
    #     ).verify()

        # AuthRule(
        #     auth_model,
        #     AuthRule(auth_model).participant_has_node_on_target(
        #         partition='ehelply-resources',
        #         node=node,
        #         target_identifier=service_target_identifier
        #     ),
        #     AuthRule(auth_model).participant_has_node_on_target(
        #         partition='ehelply-cloud',
        #         node=node,
        #         target_identifier=service_target_identifier
        #     ).entity_has_node_on_target(
        #         node=node,
        #         target_identifier=target_identifier
        #     )
        # ).verify()

        # if self.is_allowed(
        #         auth_model=auth_model,
        #         target_identifier=service_target_identifier,
        #         node=node,
        #         partition="ehelply-resources"
        # ):
        #     return True
        #
        # if self.is_allowed(
        #         auth_model=auth_model,
        #         target_identifier=auth_model.project_uuid,
        #         node=node,
        #         partition="ehelply-cloud"
        # ) and self.is_allowed(
        #     auth_model=auth_model,
        #     target_identifier=service_target_identifier,
        #     node=node,
        #     partition="ehelply-cloud"
        # ):
        #     return True
        #
        # return False



from typing import Union, Type, TypeVar

import requests

from ehelply_python_sdk.utils import SDKConfiguration, make_requests
from ehelply_python_sdk.services import services
from ehelply_python_sdk.services.service_schemas import is_response_error, ErrorResponse

genericSDKBase = TypeVar('genericSDKBase', bound=services.SDKBase)


class eHelplySDK:
    """
    eHelply SDK
    """
    def __init__(self, sdk_configuration: SDKConfiguration) -> None:
        self.sdk_configuration: SDKConfiguration = sdk_configuration
        self.requests_session: requests.Session = make_requests(sdk_configuration=sdk_configuration)

    def _make_client(
            self,
            client: str,
            sdk_configuration: SDKConfiguration = None,
            request_session: requests.Session = None
    ) -> genericSDKBase:
        if not sdk_configuration:
            sdk_configuration = self.sdk_configuration

        if not request_session:
            request_session = self.requests_session

        if client == "access":
            return services.AccessSDK(sdk_configuration=sdk_configuration, requests_session=request_session)

        if client == "security":
            return services.SecuritySDK(sdk_configuration=sdk_configuration, requests_session=request_session)

    def make_access(
            self,
            sdk_configuration: SDKConfiguration = None,
            request_session: requests.Session = None
    ) -> services.AccessSDK:
        return self._make_client(
            client="access",
            sdk_configuration=sdk_configuration,
            request_session=request_session
        )

    def make_security(
            self,
            sdk_configuration: SDKConfiguration = None,
            request_session: requests.Session = None
    ) -> services.SecuritySDK:
        return self._make_client(
            client="security",
            sdk_configuration=sdk_configuration,
            request_session=request_session
        )



from ehelply_python_sdk.services.service_schemas import *


class SearchTypeItem(BaseModel):
    uuid: str
    name: str
    slug: str
    summary: str
    created_at: str


class CreateType(BaseModel):
    name: str
    summary: str


class CreateTypeResponse(HTTPResponse):
    uuid: str
    partition_identifier: str
    name: str
    slug: str
    summary: str
    created_at: str


class SearchNodeItem(BaseModel):
    uuid: str
    name: str
    node: str
    type_uuid: str
    summary: str
    created_at: str
    roles: list


class CreateNode(BaseModel):
    name: str
    node: str
    summary: str


class CreateNodeResponse(HTTPResponse):
    uuid: str
    name: str
    node: str
    type_uuid: str
    summary: str
    created_at: str


class CreateGroup(BaseModel):
    name: str
    summary: str
    entity_identifiers: List[str] = []
    default: bool = False


class CreateGroupResponse(HTTPResponse):
    uuid: str
    partition_identifier: str
    name: str
    summary: str
    created_at: str
    default: bool


class CreateRole(BaseModel):
    name: str
    summary: str


class CreateRoleResponse(HTTPResponse):
    uuid: str
    partition_identifier: str
    name: str
    summary: str
    created_at: str


class AttachKeyToEntityResponse(HTTPResponse):
    entity_key_uuid: str
    key_uuid: str


class MakeRGTResponse(MessageResponse):
    group_uuid: str
    target_identifier: str
    role_uuid: str


class GetEntityForKeyResponse(HTTPResponse):
    entity_identifier: str



from __future__ import annotations
from typing import List, Tuple, Callable
from ehelply_python_sdk.services.access.sdk import AuthModel


class AuthException(Exception):
    pass


class AuthRule:
    """
    Provides a nice interface into developing authorization rules for endpoints
    """

    # Global config of whether to exception if unauthorized. Useful for development
    exception_if_unauthorized: bool = True

    # The exception to throw when auth fails
    exception_to_throw: Exception = AuthException

    # Global config of whether to override auth rules. Essentially, bypass authorization. Useful for development
    override: bool = False

    def __init__(
            self,
            auth_model: AuthModel,
            *rules,
            exception_if_unauthorized: bool = None,
            exception_to_throw: Exception = None,
            override: bool = None,
            execute: bool = False
    ):
        if exception_if_unauthorized is None:
            exception_if_unauthorized = AuthRule.exception_if_unauthorized
        self.local_exception_if_unauthorized: bool = exception_if_unauthorized

        if exception_to_throw is None:
            exception_to_throw = AuthRule.exception_to_throw
        self.local_exception_to_throw: Exception = exception_to_throw

        if override is None:
            override = AuthRule.override
        self.local_override: bool = override

        self.rules: List[AuthRule] = list(rules)

        self.handlers: List[Tuple[Callable, dict]] = []

        self.auth_model: AuthModel = auth_model

        if execute:
            self.verify()

    def verify(self) -> bool:
        """
        Verifies that each changed rule passes using an AND logical operation.

        If rules were passed in, it will also verify that those pass successfully. The passed in rules become a logical OR

        Returns:

        """
        rules_passed: bool = False
        for rule in self.rules:
            try:
                result: bool = rule.verify()
                if result:
                    rules_passed = True
                    break
            except:
                pass

        if not rules_passed and len(self.rules) != 0:
            if self.local_exception_if_unauthorized:
                raise self.local_exception_to_throw
            else:
                return False

        for handler in self.handlers:
            try:
                result: bool = handler[0](**handler[1])
                if not result:
                    if self.local_exception_if_unauthorized:
                        raise self.local_exception_to_throw
                    else:
                        return False
            except:
                if self.local_exception_if_unauthorized:
                    raise self.local_exception_to_throw
                else:
                    return False

        return True

    def __handler_entity_identifier_eq(self, entity_identifier: str) -> bool:
        return self.auth_model.entity_identifier == entity_identifier

    def entity_identifier_eq(self, entity_identifier: str):
        self.handlers.append((
            self.__handler_entity_identifier_eq,
            {
                "entity_identifier": entity_identifier
            }
        ))

    def __handler_entity_identifier_neq(self, entity_identifier: str) -> bool:
        return self.auth_model.entity_identifier != entity_identifier

    def entity_identifier_neq(self, entity_identifier: str):
        self.handlers.append((
            self.__handler_entity_identifier_neq,
            {
                "entity_identifier": entity_identifier
            }
        ))

    def __handler_entity_has_node_on_target(self, node: str, target_identifier: str, partition: str) -> bool:
        return self.auth_model.access_sdk.is_allowed(
            auth_model=self.auth_model,
            target_identifier=target_identifier,
            node=node,
            partition=partition
        )

    def entity_has_node_on_target(self, node: str, target_identifier: str, partition: str = None) -> AuthRule:
        self.handlers.append((
            self.__handler_entity_has_node_on_target,
            {
                "node": node,
                "target_identifier": target_identifier,
                "partition": partition
            }
        ))
        return self

    def __handler_has_entity(self) -> bool:
        return self.auth_model.entity_identifier is not None

    def has_entity(self) -> AuthRule:
        self.handlers.append((
            self.__handler_has_entity,
            {}
        ))
        return self

    def __handler_has_participant(self) -> bool:
        return self.auth_model.active_participant_uuid is not None

    def has_participant(self) -> AuthRule:
        self.handlers.append((
            self.__handler_has_participant,
            {}
        ))
        return self

    def __handler_participant_has_node_on_target(self, node: str, target_identifier: str, partition: str) -> bool:
        temp_model: AuthModel = AuthModel(
            access_sdk=self.auth_model.access_sdk,
            active_participant_uuid=self.auth_model.active_participant_uuid,
            entity_identifier=self.auth_model.active_participant_uuid,
            project_uuid=self.auth_model.project_uuid,
            access_token=self.auth_model.access_token,
            secret_token=self.auth_model.secret_token,
            claims=self.auth_model.claims,
        )

        return self.auth_model.access_sdk.is_allowed(
            auth_model=temp_model,
            target_identifier=target_identifier,
            node=node,
            partition=partition
        )

    def participant_has_node_on_target(self, node: str, target_identifier: str, partition: str = None) -> AuthRule:
        self.handlers.append((
            self.__handler_participant_has_node_on_target,
            {
                "node": node,
                "target_identifier": target_identifier,
                "partition": partition
            }
        ))
        return self

    def __handler_participant_below_limit(self) -> bool:
        return True

    def participant_below_limit(self, limit: str) -> AuthRule:
        self.handlers.append((
            self.__handler_participant_below_limit,
            {
                "limit": limit
            }
        ))
        return self

    def __handler_customentity_has_node_on_target(
            self,
            node: str,
            target_identifier: str,
            partition: str,
            entity_identifier: str
    ) -> bool:
        temp_model: AuthModel = AuthModel(
            access_sdk=self.auth_model.access_sdk,
            active_participant_uuid=self.auth_model.active_participant_uuid,
            entity_identifier=entity_identifier,
            project_uuid=self.auth_model.project_uuid,
            access_token=self.auth_model.access_token,
            secret_token=self.auth_model.secret_token,
            claims=self.auth_model.claims,
        )

        return self.auth_model.access_sdk.is_allowed(
            auth_model=temp_model,
            target_identifier=target_identifier,
            node=node,
            partition=partition
        )

    def customentity_has_node_on_target(
            self,
            node: str,
            target_identifier: str,
            partition: str,
            entity_identifier: str
    ) -> AuthRule:
        self.handlers.append((
            self.__handler_participant_has_node_on_target,
            {
                "node": node,
                "target_identifier": target_identifier,
                "partition": partition,
                "entity_identifier": entity_identifier
            }
        ))
        return self
